/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Mateusz Moscicki <m.moscicki2@partner.samsung.com>
 */

#include <gio/gio.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/procfs.h>
#include "shared/log.h"
#include "dbus_notify.h"

#define KERNEL_DEFINED_TASK_COMM_LEN 16 // from include/linux/sched.h

#define ARM_REG_LR 14
#define ARM_REG_PC 15
#define AARCH64_REG_LR 30

static int _get_important_registers(int fd, struct RegInfo **reg_info)
{
	int count = -1;

	if (reg_info == NULL) {
		_E("reg_info is NULL\n");
		return -1;
	}

	*reg_info = NULL;
	struct elf_prstatus *prstatus = mmap(NULL, sizeof(struct elf_prstatus),
					 PROT_READ, MAP_SHARED, fd, 0);
	if (prstatus == MAP_FAILED) {
		_E("mmap error: %m\n");
		return -1;
	}

#if (defined __i386__) || (defined __x86_64__)
	count = 1;
#elif (defined __arm__) || (defined __aarch64__)
	count = 2;
#endif

	*reg_info = (struct RegInfo*)malloc(sizeof(struct RegInfo)*count);
	struct RegInfo *reginfo = *reg_info;

	if (*reg_info == NULL) {
		_E("malloc for reg_info error: %m");
		goto out;
	}

#if (defined __i386__) || (defined __x86_64__)
	struct user_regs_struct *g_registers = (struct user_regs_struct*)prstatus->pr_reg;

#if defined(__i386__)
	reginfo[0].name = strdup("x86.eip");
	reginfo[0].value = (long long int)g_registers->eip;
#else
	reginfo[0].name = strdup("x86_64.rip");
	reginfo[0].value = (long long int)g_registers->rip;
#endif // defined(__i386__)

	if (reginfo[0].name == NULL)
		goto strdup_error;

#elif defined(__arm__)

	struct user_regs *g_registers;
	g_registers = (struct user_regs*)prstatus->pr_reg;

	if ((reginfo[0].name = strdup("arm.pc")) == NULL)
		goto strdup_error;
	reginfo[0].value = (long long int)g_registers->uregs[ARM_REG_PC];
	if ((reginfo[1].name = strdup("arm.lr")) == NULL)
		goto strdup_error;
	reginfo[1].value = (long long int)g_registers->uregs[ARM_REG_LR];

#elif defined(__aarch64__)

	struct user_regs_struct *g_registers = (struct user_regs_struct*)prstatus->pr_reg;

	if ((reginfo[0].name = strdup("aarch64.pc")) == NULL)
		goto strdup_error;
	reginfo[0].value = (long long int)g_registers->pc;
	if ((reginfo[1].name = strdup("aarch64.lr")) == NULL)
		goto strdup_error;
	reginfo[1].value = (long long int)g_registers->regs[AARCH64_REG_LR];

#endif // (defined __i386__) || (defined __x86_64__)

out:
	if (count <= 0) {
		free(*reg_info);
		*reg_info = NULL;
	}

	munmap(prstatus, sizeof(*prstatus));
	return count;

strdup_error:
	_E("strdup error: %m\n");
	count = -1;
	goto out;
}

static GVariant* build_message_data(const struct NotifyParams *notify_params)
{
	if (notify_params == NULL)
		return NULL;

	GVariantBuilder md_builder;
	g_variant_builder_init(&md_builder, G_VARIANT_TYPE("(sssssiia{sv})"));

	g_variant_builder_add(&md_builder, "s", notify_params->cmd_name);
	g_variant_builder_add(&md_builder, "s", notify_params->cmd_path);
	g_variant_builder_add(&md_builder, "s", notify_params->appid);
	g_variant_builder_add(&md_builder, "s", notify_params->pkgid);
	g_variant_builder_add(&md_builder, "s", notify_params->report_path);
	g_variant_builder_add(&md_builder, "i", notify_params->pid);
	g_variant_builder_add(&md_builder, "i", notify_params->tid);

	g_variant_builder_open(&md_builder, G_VARIANT_TYPE("a{sv}"));

	struct RegInfo *reg_info;
	int regs_count = _get_important_registers(notify_params->prstatus_fd, &reg_info);

	/* registers */
	for (int i = 0; i < regs_count; i++) {
		g_variant_builder_open(&md_builder, G_VARIANT_TYPE("{sv}"));

		g_variant_builder_add(&md_builder, "s", reg_info[i].name);
#if (defined __x86_64__) || (defined __aarch64__)
		g_variant_builder_add(&md_builder, "v", g_variant_new_uint64(reg_info[i].value));
#else
		g_variant_builder_add(&md_builder, "v", g_variant_new_uint32(reg_info[i].value));
#endif

		free(reg_info[i].name);
		g_variant_builder_close(&md_builder);
	}
	free(reg_info);

	g_variant_builder_close(&md_builder);

	return g_variant_builder_end(&md_builder);
}

int send_notify(const struct NotifyParams *notify_params)
{
	int result = 1;

	GDBusConnection *conn = NULL;
	GError *error = NULL;

	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
	if (error) {
		_E("g_bus_get_sync error: %s\n", error->message);
		g_error_free(error);
		result = 0;
		goto end;
	}

	GVariant *data = build_message_data(notify_params);

	if (data == NULL) {
		_E("build_message error\n");
		result = 0;
		goto end;
	}

	g_dbus_connection_emit_signal(conn,
				 NULL,
				 CRASH_PATH_CRASH,
				 CRASH_INTERFACE_CRASH,
				 PROCESS_CRASHED,
				 data,
				 &error);

	if (error) {
		_E("g_dbus_connection_emit_signal error: %s\n", error->message);
		g_error_free(error);
		result = 0;
		goto end;
	}

	g_dbus_connection_flush_sync(conn, NULL, &error);
	if (error) {
		_E("g_dbus_connection_flush_sync error: %s\n", error->message);
		g_error_free(error);
	}
end:
	if (conn != NULL)
		g_object_unref(conn);

	return result;
}
