/*
 * dump_systemstate
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file    dump_systemstate.c
 * @brief   dump system states.
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <limits.h>
#include <getopt.h>
#include <sys/stat.h>
#include <sys/vfs.h>

#include "shared/util.h"
#include "shared/log.h"

#define FILE_PERM (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)

#define TIMEOUT_DEFAULT 60

static struct dump_item {
	const char *title;
	const char *path;
} dump_item[] = {
	{"==== Binary version "            , "/etc/info.ini"},
	{"==== Tizen version "             , "/etc/tizen-release"},
	{"==== Kernel version "            , "/proc/version"},
	{"==== Boot arguments "            , "/proc/cmdline"},
	{"==== CPU & system architecture " , "/proc/cpuinfo"},
	{"==== System uptime "             , "/proc/uptime"},
	{"==== System statistics "         , "/proc/stat"},
	{"==== System memory usage "       , "/proc/meminfo"},
	{"==== Device major numbers "      , "/proc/devices"},
	{"==== System disk I/O satistics " , "/proc/diskstats"},
};

static void usage()
{
	fprintf(stderr, "usage: dump_systemstate [-k] [-d] [-f file]\n"
			"  -f: write to file (instead of stdout)\n"
			"  -k: dump kernel messages (only root)\n"
			"  -d: dump dlog messages\n"
			"  -j: dump journal log messages\n"
	   );
}

/* get disk used percentage */
static int get_disk_used_percent(const char *path)
{
	struct statfs lstatfs;
	int percent;

	if (!path)
		return -1;

	if (statfs(path, &lstatfs) < 0)
		return -1;
	percent = (((lstatfs.f_blocks - lstatfs.f_bfree) * 1000) / (lstatfs.f_blocks)) + 9;
	percent = percent/10;
	return percent;
}

int main(int argc, char *argv[])
{
#define INFORM_IF_ERROR(ret) \
	if (ret < 0) { \
		exit_code = ret; \
		fprintf_fd(out_fd, "\nCommand failed with error code: %d", ret); \
	}

	int c, ret, i, is_root, dpercent, exit_code = 0;
	const char *arg_file = NULL;
	int out_fd = -1;
	bool arg_dlog = false;
	bool arg_dmesg = false;
	bool arg_journal = false;
	char timestr[80];
	time_t cur_time;
	struct tm gm_tm;
	struct tm loc_tm;

	while ((c = getopt(argc, argv, "hf:kdj")) != -1) {
		switch (c) {
		case 'd':
			arg_dlog = true;
			break;
		case 'k':
			arg_dmesg = true;
			break;
		case 'j':
			arg_journal = true;
			break;
		case 'f':
			arg_file = optarg;
			break;
		case '?':
		case 'h':
			printf("\n");
			usage();
			ret = 0;
			goto exit;
		}
	}
	ret = 0;
	cur_time = time(NULL);
	gmtime_r(&cur_time, &gm_tm);
	localtime_r(&cur_time, &loc_tm);
	is_root = !(geteuid());

	/* open output file */
	if (arg_file == NULL) {
		out_fd = STDOUT_FILENO;
	} else {
		out_fd = open(arg_file, O_WRONLY | O_TRUNC | O_CREAT, FILE_PERM);
		if (out_fd < 0) {
			perror("couldn't open output file");
			exit_code = out_fd;
			goto exit;
		}
	}
	/* print timestamp */
	strftime(timestr, sizeof(timestr),
					"%Y-%m-%d %H:%M:%S%z", &loc_tm);
	fprintf_fd(out_fd, "dump_systemstate: %s\n", timestr);

	for (i = 0; i < ARRAY_SIZE(dump_item); i++) {
		fsync(out_fd);
		fprintf_fd(out_fd, "\n%s(%s)\n",
						dump_item[i].title, dump_item[i].path);
		ret = dump_file_write_fd(out_fd, (char *)dump_item[i].path);
		INFORM_IF_ERROR(ret)
	}
	fprintf_fd(out_fd, "\n");

	fprintf_fd(out_fd, "\n==== System disk space usage (/bin/df -h)\n");
	char *df_args[] = {"/bin/df", "-h", NULL};
	ret = run_command_write_fd_timeout(df_args[0], df_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
	INFORM_IF_ERROR(ret)

	dpercent = get_disk_used_percent("/opt");
	if (90 < dpercent) {
		fprintf_fd(out_fd, "\n==== System disk space usage detail - %d%% (/bin/du -ah /opt)\n", dpercent);
		char *du_args[] = {"/bin/du", "-ah", "/opt", "--exclude=/opt/usr", NULL};
		ret = run_command_write_fd_timeout(du_args[0], du_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
		INFORM_IF_ERROR(ret)
	}
	fprintf_fd(out_fd, "\n==== System timezone (ls -al /opt/etc/localtime)\n");
	char *ls_args[] = {"/bin/ls", "-al", "/opt/etc/localtime", NULL};
	ret = run_command_write_fd_timeout(ls_args[0], ls_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
	INFORM_IF_ERROR(ret)

	fprintf_fd(out_fd, "\n==== System summary (/usr/bin/top -bcH -n 1)\n");
	char *top_args[] = {"/bin/top", "-bcH", "-n", "1", NULL};
	char *top_env[] = {"COLUMNS=200", NULL};
	ret = run_command_write_fd_timeout(top_args[0], top_args, top_env, out_fd, NULL, 0, TIMEOUT_DEFAULT);
	INFORM_IF_ERROR(ret)

	fprintf_fd(out_fd, "\n==== Current processes (/bin/ps auxfw)\n");
	char *ps_args[] = {"/bin/ps", "auxfw", NULL};
	ret = run_command_write_fd_timeout(ps_args[0], ps_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
	INFORM_IF_ERROR(ret)

	fprintf_fd(out_fd, "\n==== System memory statistics (/usr/bin/memps -v)\n");
	char *memps_args[] = {"/bin/memps", "-v", NULL};
	ret = run_command_write_fd_timeout(memps_args[0], memps_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
	INFORM_IF_ERROR(ret)

	if (is_root) {
		fprintf_fd(out_fd, "\n==== System configuration (/usr/bin/buxton2ctl dump memory, system)\n");
		char *get_mem_args[] = {"/usr/bin/buxton2ctl", "dump", "memory", NULL};
		ret = run_command_write_fd_timeout(get_mem_args[0], get_mem_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
		INFORM_IF_ERROR(ret)

		char *get_db_args[] = {"/usr/bin/buxton2ctl", "dump", "system", NULL};
		ret = run_command_write_fd_timeout(get_db_args[0], get_db_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
		INFORM_IF_ERROR(ret)
	}

	if (arg_dmesg && is_root) {
		fprintf_fd(out_fd, "\n==== Kernel messages (TZ=UTC /bin/dmesg -T)\n");
		char *dmesg_args[] = {"/bin/dmesg", "-T", NULL};
		char *dmesg_env[] = {"TZ=UTC", NULL};
		ret = run_command_write_fd_timeout(dmesg_args[0], dmesg_args, dmesg_env, out_fd, NULL, 0, TIMEOUT_DEFAULT);
		INFORM_IF_ERROR(ret)
	}

	if (arg_dlog) {
		fprintf_fd(out_fd, "\n==== Log messages\n");
		char *dlogutil_args[] = {"/bin/dlogutil", "-d", "-v", "threadtime", "-u", "16384", NULL};
		ret = run_command_write_fd_timeout(dlogutil_args[0], dlogutil_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
		INFORM_IF_ERROR(ret)
	}

	if (arg_journal) {
		fprintf_fd(out_fd, "\n==== Journal messages\n");
		char *journalctl_args[] = {"/bin/journalctl", "-b", "-n", "1024", NULL};
		ret = run_command_write_fd_timeout(journalctl_args[0], journalctl_args, NULL, out_fd, NULL, 0, TIMEOUT_DEFAULT);
		INFORM_IF_ERROR(ret)
	}

	if (arg_file)
		close(out_fd);
exit:
	return exit_code;
#undef INFORM_IF_ERROR
}
