/*
 * crash-manager
 * Copyright (c) 2012-2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/sendfile.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <wait.h>
#include <ctype.h>
#include <grp.h>

#include "util.h"
#include "log.h"

#define READ_BUFF_SIZE 4096
#define SELECT_TIMEOUT_US 100000

int system_command_parallel(char *command)
{
	int pid = 0;
	const char *environ[] = { NULL };

	if (command == NULL)
		return -1;
	pid = fork();
	if (pid == -1)
		return -1;
	if (pid == 0) {
		char *argv[4];
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = (char *)command;
		argv[3] = 0;
		execve("/bin/sh", argv, (char **)environ);
		exit(127);
	}

	return pid;
}

int wait_system_command(int pid)
{
	int status = 0;

	if (pid < 0)
		return -1;

	do {
		if (waitpid(pid, &status, 0) == -1) {
			if (errno != EINTR)
				return -1;
		} else {
			if (WIFEXITED(status))
				return WEXITSTATUS(status);
			else if (WIFSIGNALED(status))
				return WTERMSIG(status);
			else if (WIFSTOPPED(status))
				return WSTOPSIG(status);
		}
	} while (!WIFEXITED(status) && !WIFSIGNALED(status));

	return 0;
}

int system_command(char *command)
{
	int pid = 0;

	pid = system_command_parallel(command);

	return wait_system_command(pid);
}

int system_command_with_timeout(int timeout_seconds, char *command)
{
	const char *environ[] = { NULL };

	if (command == NULL)
		return -1;
	clock_t start = clock();
	pid_t pid = fork();
	/* handle error case */
	if (pid < 0) {
		_E("fork: %d\n", errno);
		return pid;
	}
	/* handle child case */
	if (pid == 0) {
		char *argv[4];
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = (char *)command;
		argv[3] = 0;

		execve("/bin/sh", argv, (char **)environ);
		_SI("exec(%s): %d\n", command, errno);
		_exit(-1);
	}
	/* handle parent case */
	for (;;) {
		int status;
		pid_t p = waitpid(pid, &status, WNOHANG);
		float elapsed = (float) (clock() - start) / CLOCKS_PER_SEC;
		if (p == pid) {
			if (WIFSIGNALED(status))
				_SI("%s: Killed by signal %d\n", command, WTERMSIG(status));
			else if (WIFEXITED(status) && WEXITSTATUS(status) > 0)
				_SI("%s: Exit code %d\n", command, WEXITSTATUS(status));
			return WEXITSTATUS(status);
		}
		if (timeout_seconds && elapsed > timeout_seconds) {
			_SI("%s: Timed out after %.1fs (killing pid %d)\n",
					command, elapsed, pid);
			kill(pid, SIGTERM);
			return -1;
		}
		/* poll every 0.1 sec */
		usleep(100000);
	}
}

int write_fd(int fd, const void *buf, int len)
{
	int count;
	int total;
	total = 0;
	while (len) {
		count = write(fd, buf, len);
		if (count < 0) {
			if (errno == EINTR)
				continue;
			if (total)
				return total;
			return count;
		}
		total += count;
		buf = ((const char *)buf) + count;
		len -= count;
	}
	return total;
}

static int copy_bytes_sendfile(int destfd, int srcfd, off_t *ncopied)
{
	ssize_t nsent;
	off_t total = 0;

	do {
		nsent = sendfile(destfd, srcfd, NULL, INT32_MAX);
		if (nsent > 0)
			total += nsent;
		else if (nsent < 0 && errno == EAGAIN)
			continue;
		else
			break;
	} while (1);

	if (ncopied)
		*ncopied = total;

	return nsent == -1 ? -1 : 0;
}

static int copy_bytes_rw(int destfd, int srcfd, off_t *ncopied)
{
	ssize_t n;
	char buf[4096];
	off_t total = 0;

	do {
		n = read(srcfd, buf, sizeof buf);
		if (n < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;
			_E("read: %m");
			return -1;
		}

		int m = write_fd(destfd, buf, n);
		if (m != n) {
			_E("failed to write data to destination fd: %m");
			return -1;
		}
		total += m;
	} while (n > 0);

	if (ncopied)
		*ncopied = total;

	return 0;
}

int copy_bytes(int destfd, int srcfd, off_t *ncopied)
{
	int r = copy_bytes_sendfile(destfd, srcfd, ncopied);
	return r == 0 ? r : copy_bytes_rw(destfd, srcfd, ncopied);
}

int copy_file(char *dst, char *src)
{
	int res;
	int sfd;
	int dfd;

	if (!src || !dst) {
		_E("Invalid argument\n");
		return -1;
	}
	sfd = open(src, O_RDONLY);
	if (sfd < 0) {
		_E("Failed to open (%s)\n", src);
		return -1;
	}
	dfd = open(dst, O_WRONLY|O_CREAT|O_EXCL, 0644);
	if (dfd < 0) {
		close(sfd);
		_SE("Failed to open (%s)\n", dst);
		return -1;
	}

	res = copy_bytes(dfd, sfd, NULL);

	close(sfd);
	close(dfd);
	return res;
}

int move_file(char *dst, char *src)
{
	if (copy_file(dst, src) < 0)
		return -1;
	if (unlink(src) < 0)
		return -1;
	return 1;
}

int dump_file_write_fd(int dfd, char *src)
{
	int res;
	int sfd;

	if (!src) {
		_E("Invalid argument\n");
		return -1;
	}
	sfd = open(src, O_RDONLY);
	if (sfd < 0) {
		_SE("Failed to open (%s)\n", src);
		return -1;
	}

	res = copy_bytes(dfd, sfd, NULL);

	close(sfd);
	return res;
}

static int run_command(char *path, char *args[], char *env[], int fd[])
{
	if (dup2(fd[1], STDOUT_FILENO) == -1) {
		_E("dup2 error: %m");
		return -1;
	}

	if (close(fd[1]) == -1) {
		_E("close fd error: %m");
		return -1;
	}

	if (close(fd[0]) == -1) {
		_E("close fd error: %m");
		return -1;
	}

	if (execvpe(path, args, env) == -1) {
		_E("run command %s error: %m", path);
		return -1;
	}
	return -1;
}

static int wait_for_child(pid_t pid, int *exit_code, int timeout)
{
	for (int i = 0; i < 10*timeout; i++) {
		int status;
		pid_t p = waitpid(pid, &status, WNOHANG);
		if (p == pid) {
			if (WIFSIGNALED(status)) {
				_I("Killed by signal %d\n", WTERMSIG(status));
				return -1;
			} else if (WIFEXITED(status)) {
				*exit_code = WEXITSTATUS(status);
				return 0;
			}
		} else if (p == -1) {
			_E("waitpid error: %m");
			return -1;
		}
		usleep(100000);
	}
	return -1;
}

static int read_into_buff(int fd, char *buff, int size, int timeout_us, int *eof)
{
	struct timeval tout;
	int sel_ret;
	fd_set set;

	FD_ZERO(&set);
	FD_SET(fd, &set);

	tout.tv_sec = timeout_us / 1000000;
	tout.tv_usec = timeout_us % 1000000;
	*eof = 0;

	int buff_pos = 0;
	if ((sel_ret = select(fd+1, &set, NULL, NULL, &tout)) >= 0) {
		if (sel_ret > 0) {
			// we can do nonblocking read
			int readed = read(fd, &buff[buff_pos], size);

			if (readed > 0) {
				buff_pos += readed;
				size -= readed;
			} else if (readed == 0) {
				// no more data to read
				*eof = 1;
			} else {
				// error
				_E("read data from the pipe error: %m");
				return -1;
			}
		}
	} else
		_E("select() error: %m");
	return buff_pos;
}

// Usage:
// if buff is not NULL then 'size' bytes of the result is written the buffer,
// otherwise result is written to the dfd descriptor
int run_command_write_fd_timeout(char *path, char *args[], char *env[], int dfd, char *buff, int size, int timeout)
{
	char BUFF[READ_BUFF_SIZE];
	int fd[2];
	struct timeval start, end;
	int write_to_fd = buff == NULL ? 1 : 0;

	if (!write_to_fd && size <= 0) {
		_E("buffer size must be greather than zero");
		return -1;
	}

	if (pipe(fd)) {
		_E("pipe create error: %m");
		return -1;
	}

	pid_t pid = fork();

	if (pid == 0) {
		return run_command(path, args, env, fd);
	} else if (pid > 0) {
		if (close(fd[1]) == -1) {
			_E("close fd error: %m");
			return -1;
		}

		if (gettimeofday(&start, NULL) == -1) {
			_E("gettimeofday error: %m");
			return -1;
		}

		int readed;
		int eof = 0;
		int outdated = 0;
		int count = 0;

		int act_size;
		char *act_buff;

		if (write_to_fd) {
			act_size = READ_BUFF_SIZE;
			act_buff = BUFF;
		} else {
			act_size = size;
			act_buff = buff;
		}

		while ((readed = read_into_buff(fd[0], act_buff, act_size, 1000000, &eof)) >= 0) {
			if (readed > 0) {
				// we have some data
				if (count < (INT_MAX - readed))
					count += readed;
				else
					count = INT_MAX;

				if (write_to_fd) {
					if (write_fd(dfd, act_buff, readed) == -1) {
						_E("write data to pipe error: %m");
						break;
					}
				} else {
					act_buff += readed;
					act_size -= readed;

					if (act_size == 0) {
						// buff is full, we can return
						eof = 1;
					}
				}
			}

			if (eof)
				break;

			if (gettimeofday(&end, NULL) == -1) {
				_E("gettimeofday error: %m");
				break;
			}

			if ((end.tv_sec - start.tv_sec) > timeout) {
				outdated = 1;
				break;
			}

			if (readed == 0)
				usleep(100000);
		}

		if (outdated) {
			_E("command timeout: %s", path);
			if (kill(pid, 0) == 0) {
				// we can kill a child because we don't
				// need it anymore
				if (kill(pid, SIGTERM) == -1)
					_E("kill child %d error: %m", pid);
			}
		}

		if (close(fd[0]) == -1) {
			_E("close fd error: %m");
			return -1;
		}

		// let's wait a second for a child
		int exit_code = -1;
		int wait_res = wait_for_child(pid, &exit_code, 1);

		if (wait_res != 0)
			_I("wait_for_child for \%s\" returns non-zero value\n", path);
		else if (exit_code != 0)
			_I("\"%s\" exit code: %d\n", path, exit_code);

		return (eof == 1 && exit_code == 0) ? count : -abs(exit_code);
	} else {
		_E("fork() error: %m");
		return -1;
	}

	return -1;
}

int run_command_timeout(char *path, char *args[], char *env[], int timeout)
{
	int fd = open("/dev/null", O_WRONLY);
	if (fd < 0) {
		_E("open /dev/null error: %m");
		return -1;
	}

	int res = run_command_write_fd_timeout(path, args, env, fd, NULL, 0, timeout);

	close(fd);

	return res;
}

int fsync_path(char *const path)
{
	int fd, ret;

	ret = fd = open(path, O_RDONLY);
	if (fd >= 0) {
		ret = fsync(fd);
		close(fd);
	}

	if (ret < 0)
		_E("Unable to fsync %s: %m", path);

	return ret;
}

static int remove_dir_internal(int fd)
{
	DIR *dir;
	struct dirent *de;
	int subfd, ret = 0;

	dir = fdopendir(fd);
	if (!dir)
		return -1;

	while ((de = readdir(dir))) {
		if (de->d_type == DT_DIR) {
			if (!strncmp(de->d_name, ".", 2) || !strncmp(de->d_name, "..", 3))
				continue;
			subfd = openat(fd, de->d_name, O_RDONLY | O_DIRECTORY);
			if (subfd < 0) {
				_SE("Couldn't openat %s: %d\n", de->d_name, errno);
				ret = -1;
				continue;
			}
			if (remove_dir_internal(subfd))
				ret = -1;
			close(subfd);
			if (unlinkat(fd, de->d_name, AT_REMOVEDIR) < 0) {
				_SE("Couldn't unlinkat %s: %d\n", de->d_name, errno);
				ret = -1;
			}
		} else {
			if (unlinkat(fd, de->d_name, 0) < 0) {
				_SE("Couldn't unlinkat %s: %d\n", de->d_name, errno);
				ret = -1;
			}
		}
	}
	closedir(dir);
	return ret;
}

int remove_dir(const char *path, int del_dir)
{
	int fd, ret = 0;

	if (!path)
		return -1;
	fd = open(path, O_RDONLY | O_NONBLOCK | O_DIRECTORY | O_CLOEXEC | O_NOFOLLOW);
	if (fd < 0) {
		_SE("Couldn't opendir %s: %d\n", path, errno);
		return -errno;
	}
	ret = remove_dir_internal(fd);
	close(fd);

	if (del_dir) {
		if (rmdir(path)) {
			_SE("Couldn't rmdir %s: %d\n", path, errno);
			ret = -1;
		}
	}
	return ret;
}

int get_exec_pid(const char *execpath)
{
	DIR *dp;
	struct dirent *dentry;
	int pid = -1, fd;
	int ret;
	char buf[PATH_MAX];
	char buf2[PATH_MAX];
	int len;

	dp = opendir("/proc");
	if (!dp) {
		_E("FAIL: open /proc");
		return -1;
	}

	len = strlen(execpath) + 1;

	while ((dentry = readdir(dp))) {
		if (!isdigit(dentry->d_name[0]))
			continue;

		pid = atoi(dentry->d_name);

		snprintf(buf, PATH_MAX, "/proc/%d/cmdline", pid);
		fd = open(buf, O_RDONLY);
		if (fd < 0)
			continue;
		ret = read(fd, buf2, PATH_MAX);
		close(fd);

		if (ret < 0 || ret >= PATH_MAX)
			continue;

		buf2[ret] = '\0';

		if (!strncmp(buf2, execpath, len)) {
			closedir(dp);
			return pid;
		}
	}

	errno = ESRCH;
	closedir(dp);
	return -1;
}

int get_file_count(char *path)
{
	DIR *dir;
	struct dirent *dp;
	int count = 0;

	dir = opendir(path);
	if (!dir)
		return 0;

	while ((dp = readdir(dir))) {
		const char *name = dp->d_name;
		/* always skip "." and ".." */
		if (name[0] == '.') {
			if (name[1] == 0)
				continue;
			if ((name[1] == '.') && (name[2] == 0))
				continue;
		}
		count++;
	}
	closedir(dir);
	return count;
}

off_t get_directory_usage(char *path)
{
	DIR *dir;
	struct dirent *de;
	struct stat st;
	off_t usage = 0;
	int fd = -1;

	fd = open(path, O_RDONLY|O_NONBLOCK|O_DIRECTORY|O_CLOEXEC|O_NOFOLLOW|O_NOATIME);
	if (fd < 0)
		return -1;
	dir = fdopendir(fd);
	if (!dir) {
		close(fd);
		return -1;
	}

	while ((de = readdir(dir))) {
		if (!strncmp(de->d_name, ".", 2) || !strncmp(de->d_name, "..", 3))
			continue;
		if (fstatat(fd, de->d_name, &st, AT_SYMLINK_NOFOLLOW) < 0) {
			_SE("Failed to fstatat  %s: %d\n", de->d_name, errno);
			continue;
		}
		usage += st.st_size;
	}
	closedir(dir);
	close(fd);
	return usage;
}

int log_kmsg(char *fmt, ...)
{
	int result = 1;
	va_list ap;
	FILE *file;

	file = fopen("/dev/kmsg", "w");
	if (file == NULL) {
		_E("Open /dev/kmsg error: %m");
		return 0;
	}

	va_start(ap, fmt);
	if (vfprintf(file, fmt, ap) < 0) {
		_E("Write to /dev/kmsg error: %m");
		result = 0;
	}
	va_end(ap);
	fclose(file);
	return result;
}

/**
 * @brief Check wchan of thread
 *
 * @param pid PID of the inspected process
 * @param tid TID of the thread to check
 */
static int check_thread_wchan(int pid, int tid)
{
	int fd, cnt;
	char path[PATH_MAX], buf[100];

	snprintf(path, sizeof(path), "/proc/%d/task/%d/wchan", pid, tid);
	fd = open(path, O_RDONLY);
	if (fd == -1) {
		_E("cannot open %s: %m\n", path);
		return -errno;
	}
	cnt = read(fd, buf, sizeof(buf));
	if (cnt == -1 || cnt == sizeof(buf)) {
		_E("read %s error: %m\n", path);
		close(fd);
		return -errno;
	}
	buf[cnt] = 0;
	close(fd);

	if (strncmp("do_coredump", buf, sizeof(buf)) == 0 || strncmp("pipe_wait", buf, sizeof(buf)) == 0)
		return tid;
	else
		return 0;
}

/**
 * @brief Find crashed tid if tid was not offered
 *
 * @param pid PID of the inspected process
 */
int find_crash_tid(int pid)
{
	int threadnum = 1;
	int crash_tid = -1;
	DIR *dir;
	struct dirent *entry;
	char task_path[PATH_MAX];
	struct stat sb;

	snprintf(task_path, sizeof(task_path), "/proc/%d/task", pid);
	if (stat(task_path, &sb) == -1) {
		_E("no such file: %s", task_path);
		return -1;
	}

	threadnum = sb.st_nlink - 2;

	if (threadnum > 1) {
		dir = opendir(task_path);
		if (!dir) {
			_E("cannot open %s\n", task_path);
			return -1;
		} else {
			while ((entry = readdir(dir)) != NULL) {
				if (strcmp(entry->d_name, ".") == 0 ||
				    strcmp(entry->d_name, "..") == 0)
					continue;
				crash_tid = check_thread_wchan(pid,
							       atoi(entry->d_name));
				if (crash_tid > 0)
					break;
			}
			closedir(dir);
			return crash_tid;
		}
	} else if (threadnum == 1) {
		return pid;
	}
	return -1;
}

char* get_cmd_line(pid_t pid)
{
	char cmdline_path[PATH_MAX];

	snprintf(cmdline_path, sizeof(cmdline_path),
			"/proc/%d/cmdline", pid);

	int fd = open(cmdline_path, O_RDONLY);
	if (fd < 0) {
		_E("Failed to open %s: %m\n", cmdline_path);
		return NULL;
	}

	char buffer[PATH_MAX];
	ssize_t ret = read(fd, buffer, sizeof(buffer) - 1);
	close(fd);

	if (ret <= 0) {
		_E("Failed to read %s: %m\n", cmdline_path);
		return NULL;
	}
	buffer[ret] = '\0';

	char *result;
	if (asprintf(&result, "%s", buffer) == -1) {
		_E("asprintf() error: %m\n");
		return NULL;
	}

	return result;
}

char* get_exe_path(pid_t pid)
{
	char exe_link[PATH_MAX];
	char buffer[PATH_MAX];

	snprintf(exe_link, sizeof(exe_link),
		"/proc/%d/exe", pid);

	ssize_t ret = readlink(exe_link, buffer, sizeof(buffer) - 1);

	if (ret <= 0) {
		_E("Failed to read link %s: %m", exe_link);
		return NULL;
	}
	buffer[ret] = '\0';

	if (access(buffer, F_OK) == -1) {
		_E("Invalid path %s", buffer);
		return NULL;
	}

	char *result;
	if (asprintf(&result, "%s", buffer) == -1) {
		_E("asprintf() error: %m\n");
		return NULL;
	}

	return result;
}

/* This function is supposed to accept same data as passed to execve
 * (argv and envp), which can be arrays of strings as well as NULL
 * pointer.
 */
char* concatenate(char *const vec[])
{
	size_t length = 0;
	for (char *const *p = vec; p && *p; p++)
		length += strlen(*p) + 1;

	if (length == 0)
		return strdup("");

	char *str = (char *)malloc(length);
	if (!str)
		return NULL;

	char *destp = str;
	char *const *vecp = vec;
	while (*vecp) {
		destp = stpcpy(destp, *(vecp++));
		if (*vecp)
			destp = stpcpy(destp, " ");
	}

	return str;
}

/**
 * @}
 */
